from os import environ
from os.path import expanduser
from pathlib import Path


def get_app_config(app):
    path = Path(f'~/.config/{app}').expanduser()
    try:
        from pybrary.modules import load
        full = path / 'config.py'
        config = load('config', full)
        return full, config.config
    except: pass
    try:
        from tomllib import loads
        full = path / 'config.toml'
        config = loads(full.read_text())
        return full, config
    except: pass
    try:
        from yaml import load, SafeLoader
        full = path / 'config.yaml'
        config = load(full, loader=SafeLoader)
        return full, config
    except: pass
    try:
        from json import load
        full = path / 'config.json'
        config = load(full)
        return full, config
    except: pass
    return None, None



from base64 import urlsafe_b64encode
from hashlib import blake2b

from cryptography.fernet import Fernet as Base


class Fernet:
    def __init__(self, psw):
        key = self.make_key(psw)
        self.fernet = Base(key)

    @staticmethod
    def make_key(psw):
        h = blake2b(digest_size=32)
        h.update(psw.encode())
        key = urlsafe_b64encode(h.digest())
        return key

    def encrypt(self, data):
        return self.fernet.encrypt(data.encode())

    def decrypt(self, data):
        return self.fernet.decrypt(data).decode()



from html.parser import HTMLParser
from urllib.request import urlopen

from . import debug

class Node:
    def __init__(self, parent, tag, attrs):
        self.parent = parent
        self.tag = tag
        self.attrs = dict(attrs)
        self.text = None
        self.children = []

    @property
    def level(self):
        return self.parent.level+1 if self.parent else 0

    def add(self, node):
        self.children.append(node)

    def __iter__(self):
        for child in self.children:
            yield child
            yield from child

    @property
    def classes(self):
        return self.attrs.get('class',  '').split(' ')

    def __getattr__(self, key):
        return self.attrs.get(key, '')

    def __str__(self):
        return f'{"  "*self.level}{self.tag} {self.classes}'


class Parser(HTMLParser):
    auto_end = 'img', 'link', 'meta'

    def handle_starttag(self, tag, attrs):
        if self.current.tag in self.auto_end:
            self.handle_endtag(self.current.tag)
        # print('  -->', tag)
        node = Node(self.current, tag, attrs)
        self.current.add(node)
        self.current = node

    def handle_data(self, data):
        self.current.text = data.strip()

    def handle_endtag(self, tag):
        current = self.current.tag
        if current!=tag and current in self.auto_end:
            self.handle_endtag(current)
        # print('<--  ' ,tag)
        self.current = self.current.parent

    def load(self, html):
        debug('parser load %s', len(html))
        self.root = Node(None, 'root', {})
        self.current = self.root
        self.feed(html)
        self.close()
        debug('parser html loaded')
        return self.root

    def get(self, url):
        html = urlopen(url).read().decode('utf-8')
        return self.load(html)



from functools import partial
from datetime import datetime
from io import SEEK_END
from operator import attrgetter
from os import scandir
import os, stat
from pathlib import Path
from re import compile, match


def files(path):
    with scandir(path) as ls:
        for entry in sorted(ls, key=attrgetter('path')):
            if entry.is_dir(follow_symlinks=False):
                yield from files(entry.path)
            elif entry.is_file():
                yield Path(entry)


def find(path, name):
    match = compile(name).search
    for path in files(path):
        if match(str(path)):
            yield path


def grep(path, rex):
    rmatch = partial(match, compile(rex))
    with open(path) as inp:
        for found in filter(None, map(rmatch, inp)):
            yield found.groups()


class FileStat:
    def __init__(self, path):
        self.path = path if isinstance(path, Path) else Path(path)
        self.stat = os.stat(str(self.path))

    @property
    def mtime(self):
        mtime = self.stat[stat.ST_MTIME]
        return datetime.fromtimestamp(mtime)

    @property
    def age(self):
        delta = datetime.now() - self.mtime
        return delta.days


def last_lines(path, nb_lines, max_line_len=99):
    offset = nb_lines * max_line_len
    with open(path, 'rb') as inp:
        try:
            inp.seek(-offset, SEEK_END)
        except OSError:
            inp.seek(0)
        raw = inp.read()
    lines = raw.decode().split('\n')
    last = lines[-nb_lines:]
    return last




from sys import exit

from pybrary.command import parse_args
from pybrary import commands


def main():
    a, k = parse_args()
    f, *a = a
    try:
        cmd = getattr(commands, f)
        success, output = cmd(*a, **k)
    except Exception as x:
        success, output = False, str(x)
    if success:
        print(output)
    else:
        exit(f' ! {output}')



from urllib.request import urlopen
from ipaddress import ip_address, IPv4Address

from .shell import shell
from .commandparams import Param, ValidationError
from . import logger, debug, error


def get_ip_adr():
    url = 'http://api.ipify.org'
    with urlopen(url, timeout=1) as resp:
        if resp.status==200:
            adr = resp.read().decode().strip()
            debug('ip adr : %s', adr)
            return True, adr
        else:
            error('%s %s', resp.status, resp.reason)
            return False, resp.reason


class ParamIPv4(Param):
    name = 'IPv4 adr'

    def verify(self, value):
        try:
            adr = ip_address(value)
            if not isinstance(adr, IPv4Address):
                raise ValueError
            return adr
        except ValueError:
            raise ValidationError(f'{value} is not IPv4')




from importlib.util import spec_from_file_location
from importlib.util import module_from_spec

from . import debug

def load(mod_name, mod_file):
    debug('module.load name:%s file:%s', mod_name, mod_file)
    spec = spec_from_file_location(mod_name, mod_file)
    mod = module_from_spec(spec)
    spec.loader.exec_module(mod)
    debug('module loaded : %s', mod)
    return mod



from pytest import raises

from pybrary.primes import primes


def test_n_contains():
    assert 0 not in primes
    assert 1 not in primes
    assert 2 in primes
    assert 3 in primes
    assert 4 not in primes
    assert 5 in primes
    assert 6 not in primes
    assert 7 in primes
    assert 8 not in primes
    assert 9 not in primes
    assert 11159 in primes


def test_n_iter():
    expected = 2,3,5,7,11,13,17,19,23
    for e, p in zip(expected, primes):
        assert p == e

def test_n_iindex():
    assert primes[0] == 2
    assert primes[1] == 3
    assert primes[2] == 5
    assert primes[3] == 7

def test_c_index_error():
    with  raises(OverflowError):
        primes[-1]

def test_n_decomp():
    assert primes.factors(12) == [2, 2, 3]
    assert primes.factors(15) == [3, 5]
    assert primes.factors(179) == [179]
    assert primes.factors(1) == []
    assert primes.factors(2**9) == [2 for _ in range(9)]
    assert primes.factors(7**7) == [7 for _ in range(7)]


if __name__=='__main__':
    from pytest import main
    main(['-v', '-s'])



from functools import partial
from subprocess import (
    PIPE,
    CalledProcessError,
    run,
)


def shell(script, shell):
    try:
        proc = run(
            script,
            stdout = PIPE,
            stderr = PIPE,
            check  = True,
            shell  = True,
            executable = shell,
        )
    except CalledProcessError as exc:
        ret = exc.returncode
        out = exc.stdout.decode('utf-8')
        err = exc.stderr.decode('utf-8')
    else:
        ret = proc.returncode
        out = proc.stdout.decode('utf-8').strip()
        err = proc.stderr.decode('utf-8').strip()
    return ret, out, err


bash = partial(shell, shell='/bin/bash')



from os import environ
from sys import argv


class ValidationError(Exception):
    '''Validation Error'''


class SubCmdError(Exception):
    '''Sub Command Error'''


class Param:
    name = None
    positional = False
    mandatory = False
    default = None

    def __init__(self, parent):
        if not (self.__doc__ and self.__doc__.strip()):
            raise ValueError(
                f'"{self.__class__.__name__}" Param should have a doc string.'
            )
        if not self.name:
            raise ValueError(
                f'"{self.__class__.__name__}" Param should have a "name" attribute.'
            )
        self.parent = parent # command or param
        self.get = self.parent.get
        self.set = self.parent.set

    def verify(self, value):
        return value

    def validate(self, value):
        return self.verify(value)

    def check(self, val):
        try:
            v = self.validate(val)
        except ValidationError as x:
            v = f' ! {val} !'
        return v

    def help(self, command=None):
        if command:
            value = command.get(self.name)
            print(f'\n{self.name} = {value}')
        else:
            print(f'\n{self.name}')
        self.positional and print(' - Positional')
        self.mandatory and print(' - Mandatory')
        self.default and print(f' - Default : {self.default}')
        lines = [f'    {i}' for i in self.__doc__.strip().split('\n')]
        lines[0] = f'    * {lines[0].strip()} *'
        print('\n'.join(lines))


def mk_subparam(cmd):
    class SubCmdParam(Param):
        '''SubCmd
    Sub Command
        '''
        name = 'subcmd'
        positional = True
        mandatory = False
        command = cmd.__class__

        def verify(self, value):
            return self.command.subs.get(value)

    return SubCmdParam(cmd)


def parse_args(args=None):
    if args:
        opt = args.split()
    else:
        _cmd, *opt = argv

    args, kwds = [], {}
    for o in opt:
        if '=' in o:
            k, v = o.split('=')
            kwds[k] = v
        else:
            args.append(o)
    return args, kwds


class Command:
    config = None
    env_pfx = None
    subs = None
    vargs = False
    line = False
    Params = []
    shortcut = False

    def __init__(self, *Params, parent=None):
        if not (self.__doc__ and self.__doc__.strip()):
            raise ValueError(
                f'"{self.__class__.__name__}" Command should have a doc string.'
            )
        self.parent = parent
        self._args = list()
        self._kws = dict()
        self.params = dict()
        params = [P(self) for P in Params or self.Params]
        if self.subs:
            subparam = mk_subparam(self)
            params.append(subparam)
        self.specs = {
            p.name: p
            for p in params
        }
        self.positionals = [
            p
            for p in params
            if p.positional
        ]

    @property
    def args(self):
        return list(self._args)

    @property
    def kws(self):
        return {
            k:v
            for k,v in self._kws.items()
            if not k.startswith('_')
        }

    def get(self, name):
        if name in self.params:
            return self.params[name]
        if self.parent:
            return self.parent.get(name)

    def set(self, name, value):
        self.params[name] = value

    @property
    def env_prefix(self):
        def mk(pref):
            if isinstance(pref, str) and pref.strip():
                return f'{pref.strip()}_'
        pref = (
            mk(self.env_pfx)
            or mk(self.config and self.config.get('env_prefix'))
            or ''
        )
        return pref

    def validate(self, *a, **k):
        check = k.pop('_check', False)
        pick = k.pop('_pick', False)
        params = dict()

        if self.line:
            return dict(
                _line_ = ' '.join(a)
            )

        for i, p in enumerate(list(self.positionals)):
            if p.name in k:
                del self.positionals[i]

        if len(a) >  len(self.positionals):
            if not pick:
                raise ValidationError(f'Too many args : {len(a)}. Expected at most {len(self.positionals)}')

        for v, p in zip(a, self.positionals):
            try:
                val = p.check(v) if check else p.validate(v)
            except ValidationError:
                if self.config and p.name in self.config:
                    v = self.config[p.name]
                    val = p.check(v) if check else p.validate(v)
                else:
                    if p.default is None:
                        raise
                    else:
                        v = p.default
                        val = p.check(v) if check else p.validate(v)
            else:
                if pick and val:
                    self._args.remove(v)
            params[p.name] = val

        for name, val in k.items():
            if name in params:
                raise ValidationError(f'{name} is overridding positional arg')
            p = self.specs.get(name)
            if p:
                val = p.check(val) if check else p.validate(val)
                params[name] = val
                if pick:
                    del self._kws[name]
            else:
                if not pick:
                    params[name] = '?'
                    if not check:
                        raise ValidationError(f'Unsupported arg : {name}')

        for name, p in self.specs.items():
            if self.get(name):
                continue
            if name not in params and self.config:
                if name in self.config:
                    v  = self.config[name]
                    val = p.check(v) if check else p.validate(v)
                    params[name] = val
            if name not in params:
                var = f'{self.env_prefix}{name}'
                if var in environ:
                    v = environ[var]
                    val = p.check(v) if check else p.validate(v)
                    params[name] = val
            if name not in params and p.default:
                val = p.check(p.default) if check else p.validate(p.default)
                params[p.name] = val
            if name not in params:
                if p.mandatory:
                    params[name] = '!'
                    if not check:
                        raise ValidationError(f'Missing mandatory arg : {p.name}')
                else:
                    params[name] = None

        return params

    def get_params(self, args=None, _check=False, _pick=False):
        if args:
            a, k = parse_args(args)
        elif self.parent:
            a = self.parent.args
            k = self.parent._kws
        else:
            a, k = parse_args()

        if not _check:
            self._args = a
            self._kws = k

        k['_check'] = _check
        k['_pick'] = _pick

        try:
            # match and pick
            params = self.validate(*a, **k)
        except ValidationError:
            # no match
            raise
        else:
            # match, not pick
            if self._args == list(params.values()):
                self._args = []

        return params

    def __call__(self, args=None):
        if self.subs:
            try:
                self.params = self.get_params(args, _pick=True)
            except ValidationError:
                self.help()
                raise
            subcmd = self.params.get('subcmd')
            if subcmd:
                subcmd.parent = self
                try:
                    return subcmd()
                except ValidationError as x:
                    subcmd.help()
                    print(f'\n ! {x}\n')
                    raise SubCmdError
            else:
                for name, cmd in self.subs.items():
                    try:
                        if cmd.shortcut:
                            cmd.parent = self
                            try:
                                return cmd(args)
                            except SubCmdError:
                                return
                            except ValidationError:
                                continue
                        else:
                            cmd = cmd.subs and cmd.subs.get(self.args[0])
                            if cmd:
                                cmd.parent = self
                                self._args = self._args[1:]
                                return cmd(' '.join(self.args))
                    except Exception as x:
                        print(f' ! {name} - {cmd} ! {x} ! ')

            if self.parent:
                raise ValidationError
            else:
                self.help()

        else:
            try:
                params = self.get_params(args, _pick=self.vargs)
                self.params.update(params)
            except ValidationError as x:
                if self.parent and self.shortcut:
                    raise
                try:
                    self.help(args)
                except Exception as exc:
                    print(f' ! {exc}')
                    return
                print(f' ! {x}')
                return

            result = self.run()
            self.params = dict()
            return result


    def help(self, args=None, fetch=True):
        print(self.__doc__)
        if fetch:
            params = self.get_params(args, _check=True)
            for p in self.specs.values():
                p.help(self)
        else:
            for p in self.specs.values():
                p.help()
        print()
        if self.subs:
            subs = '\n    '.join(self.subs.keys())
            print(f'Sub Commands :\n    {subs}\n')



from warnings import warn

from pybrary.func import caller


class ContractError(Exception):
    ''' Contract Error '''

class PreError(ContractError):
    ''' Contract Pre Error '''

class ResultError(ContractError):
    ''' Contract Result Error '''

class PostError(ContractError):
    ''' Contract Post Error '''


class Contract:
    def __init__(self, main):
        self.main = main
        self.name = self.__class__.__name__
        self.func = f'{main.__module__}.{main.__name__}'

    def warn(self):
        warn(f'\n\n ? "{caller()}" not defined for Contract "{self.name}"\n')

    def pre(self, *a, **k):
        self.warn()

    def alt(self, *a, **k):
        self.warn()
        return 'no_alt'

    def _run(self, *a, **k):
        r = self.main(*a, **k)
        x = self.alt(*a, **k)
        if x!='no_alt':
            aa = ', '.join(map(str, a)) if a else ''
            kk = ', '+str(k) if k else ''
            assert r == x, f'{self.name} results mismatch {self.func}({aa}{kk}) returned {r} expected {x}'
        return r

    def post(self, result, *a, **k):
        self.warn()

    def __call__(self, *a, **k):
        try:
            self.pre(*a, **k)
        except Exception as x:
            raise PreError(f'\n\n ! {x} !')
        else:
            try:
                r = self._run(*a, **k)
            except Exception as x:
                raise ResultError(f'\n\n ! {x} !')
            else:
                try:
                    self.post(r, *a, **k)
                except Exception as x:
                    raise PostError(f'\n\n ! {x} !')
                else:
                    return r


def contract(cls):
    def deco(fct):
        return cls(fct) if __debug__ else fct
    return deco



from os.path import join, dirname, abspath

from setuptools import setup, find_packages

curdir = abspath(dirname(__file__))
readme = open(join(curdir, 'README.rst')).read()

setup(
    name             = 'pybrary',
    version          = '0.22.27.0',
    description      = 'Python Library',
    long_description = readme,
    keywords         = ['library', ],
    url              = 'https://framagit.org/louis-riviere-xyz/pybrary/tree/stable',
    author           = 'Louis RIVIERE',
    author_email     = 'louis@riviere.xyz',
    license          = 'MIT',
    classifiers      = [
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        'Intended Audience :: Developers',
    ],
    packages = find_packages(),
    install_requires = [
        'cryptography',
    ],
    entry_points = dict(
        console_scripts = (
            'pybrary=pybrary.main:main',
        ),
    ),
)
